#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#
# Copyright 2012 posativ <info@posativ.org>. All rights reserved.
# License: BSD Style, 2 clauses. see acrylamid/__init__.py

import os
import re
import tempfile
import getpass

from base64 import b64encode
from datetime import datetime
from urllib2 import urlopen, Request, HTTPError
from urlparse import urlsplit

from xml.etree import ElementTree
from htmlentitydefs import name2codepoint

from email.utils import parsedate_tz, mktime_tz
from os.path import join, dirname, isfile

from acrylamid import log
from acrylamid.utils import FileEntry, event, escape, system
from acrylamid.errors import AcrylamidException


class ParseException(Exception):
    pass


def _unescape(s):
    return re.sub('&(%s);' % '|'.join(name2codepoint),
            lambda m: unichr(name2codepoint[m.group(1)]), s)


def _convert(data, fmt='markdown'):
    """Reconversion of HTML to Markdown or reStructuredText.  Defaults to Markdown,
    but can be in fact every format pandoc supports. If pandoc is not available, try
    some specific conversion tools like html2text and html2rest.

    :param html: raw content to convert to
    :param html: format to reconvert to"""

    if fmt in ('Markdown', 'markdown', 'mkdown', 'md', 'mkd'):
        cmds = ['html2text']
        fmt = 'markdown'
    elif fmt in ('rst', 'restructuredtext', 'rest', 'reStructuredText'):
        cmds = ['html2rest']
        fmt = 'rst'
    else:
        cmds = []

    cmds.insert(0, ['pandoc', '-f', 'html', '-t', fmt, '--strict'])

    if fmt == 'html':
        return data, 'html'

    #  - item.find(foo).text returns None if no CDATA
    #  - pandoc waits for input if a zero-length string is given
    if data is None or data is '':
        return '', fmt

    for cmd in cmds:
        try:
            return system(cmd, stdin=str(data.decode('utf-8'))), fmt.lower()
        except AcrylamidException as e:
            log.warn(e.message)
        except OSError:
            pass
    else:
        return data, 'html'


def _rss20(content):

    def parse_date_time(stamp):
        ts = parsedate_tz(stamp)
        ts = mktime_tz(ts)
        return datetime.fromtimestamp(ts)

    try:
        tree = ElementTree.fromstring(content)
    except ElementTree.ParseError:
        raise AcrylamidException('no well-formed XML')
    if tree.tag != 'rss' and tree.attrib.get('version') == '2.0':
        raise ParseException('no RSS 2.0 feed')

    # --- site settings --- #

    defaults = {}
    channel = tree.getchildren()[0]

    for k, v in {'title': 'sitename', 'link': 'www_root', 'language': 'lang'}.iteritems():
        try:
            defaults[v] = channel.find(k).text
        except AttributeError:
            pass

    # --- individual posts --- #

    items = []
    for item in channel.findall('item'):

        entry = {}
        for k, v in {'title': 'title', 'date': 'pubDate', 'link': 'link',
                     'content': 'description'}.iteritems():
            try:
                entry[k] = item.find(v).text if k != 'content' \
                                             else _unescape(item.find(v).text)
            except (AttributeError, TypeError):
                pass

        try:
            entry['content'] = [_.text for _ in item.getchildren() if _.tag.endswith('encoded')][0]
        except (AttributeError, IndexError):
            pass

        if filter(lambda k: not k in entry, ['title', 'date', 'link', 'content']):
            raise ParseException('invalid RSS 2.0 feed: provide at least a title, '
                                 + 'link, content and pubDate!')

        items.append({
            'title': entry['title'],
            'content': entry['content'],
            'date': parse_date_time(entry['date']),
            'link': entry['link']
            })

    return defaults, items


def fetch(url, auth=None):
    """Fetch URL, optional with HTTP Basic Authentication."""

    if not (url.startswith('http://') or url.startswith('https://')):
        try:
            with open(url, 'r') as fp:
                content = fp.read()
            return content
        except OSError as e:
            raise AcrylamidException(e.message)

    req = Request(url)
    if auth:
        req.add_header('Authorization', 'Basic ' + b64encode(auth))

    try:
        r = urlopen(req)
    except HTTPError as e:
        raise AcrylamidException(e.msg)

    if r.getcode() == 401:
        user = raw_input('Username: ')
        passwd = getpass.getpass()
        fetch(url, user + ':' + passwd)
    elif r.getcode() == 200:
        return r.read()

    raise AcrylamidException('invalid status code %i, aborting.' % r.getcode())


def parse(content):

    for method in (_rss20,):
        try:
            return method(content)
        except ValueError as e:
            raise AcrylamidException(e.message)
        except ParseException:
            pass
    else:
        raise AcrylamidException('unable to parse feed.')


def build(conf, env, defaults, items, fmt, keep=False):

    def create(title, date, content, permalink=None):
        fd, tmp = tempfile.mkstemp(suffix='.txt', dir='.cache/')
        title = escape(title)

        with os.fdopen(fd, 'wb') as f:
            f.write('---\n')
            f.write('title: %s\n' % title)
            f.write('date: %s\n' % date.strftime(conf['date_format']))
            f.write('filter: [%s, ]\n' % fmt)
            if permalink:
                f.write('permalink: %s\n' % permalink)
            f.write('---\n\n')
            f.write(content[0])

        entry = FileEntry(tmp, conf)
        p = join(conf['entries_dir'], dirname(entry.permalink)[1:])

        try:
            os.makedirs(p.rsplit('/', 1)[0])
        except OSError:
            pass

        filepath = p + '.txt'
        if isfile(filepath):
            raise AcrylamidException('Entry already exists %r' % filepath)
        os.rename(tmp, filepath)
        event.create(filepath)

    for item in items:

        if keep:
            m = urlsplit(item['link'])
            permalink = m.path if m.path != '/' else None

        create(item['title'], item['date'], _convert(item.get('content', ''), fmt),
               permalink=permalink if keep else None)



# Copyright 2012 posativ <info@posativ.org>. All rights reserved.
# License: BSD Style, 2 clauses. see acrylamid/__init__.py

from acrylamid.filters import Filter
from re import sub


class Headoffset(Filter):

    match = ['h' + str(i + 1) for i in range(5)]

    def transform(self, text, entry, *args):

        def f(m):
            '''will return html with all headers increased by 1'''
            l = lambda i: i if i not in [str(x) for x in range(1, 6)] \
                            else str(int(i) + 1) if int(i) < 5 else '6'
            return ''.join([l(i) for i in m.groups()])

        offset = int(self.name[1])

        for i in range(offset):
            text = sub('(<h)(\d)(>)(.+)(</h)(\d)(>)', f, text)

        return text



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#
# Copyright 2012 posativ <info@posativ.org>. All rights reserved.
# License: BSD Style, 2 clauses. see acrylamid/__init__.py

from acrylamid.filters import Filter
from acrylamid.utils import cached_property


class Restructuredtext(Filter):

    match = ['restructuredtext', 'rst', 'rest', 'reST', 'reStructuredText']
    conflicts = ['markdown', 'plain']
    priority = 70.00

    __ext__ = dict((x, x) for x in [])

    def transform(self, content, entry, *filters):
        initial_header_level = 1
        transform_doctitle = 0
        settings = {
            'initial_header_level': initial_header_level,
            'doctitle_xform': transform_doctitle
            }
        parts = self.publish_parts(content, writer_name='html', settings_overrides=settings)
        return parts['body'].encode('utf-8')

    @cached_property
    def publish_parts(self):
        """On-demand import.  Importing reStructuredText and Pygments takes
        3/5 of overall runtime just to __init__ and fill memory."""

        from docutils import nodes
        from docutils.core import publish_parts
        from docutils.parsers.rst import directives, Directive

        from pygments import highlight
        from pygments.formatters import HtmlFormatter
        from pygments.lexers import get_lexer_by_name, TextLexer

        # Set to True if you want inline CSS styles instead of classes
        INLINESTYLES = False

        # The default formatter
        DEFAULT = HtmlFormatter(noclasses=INLINESTYLES)

        # Add name -> formatter pairs for every variant you want to use
        VARIANTS = {
            # 'linenos': HtmlFormatter(noclasses=INLINESTYLES, linenos=True),
        }

        class Pygments(Directive):
            """ Source code syntax hightlighting.
            """
            required_arguments = 1
            optional_arguments = 0
            final_argument_whitespace = True
            option_spec = dict([(key, directives.flag) for key in VARIANTS])
            has_content = True

            def run(self):
                self.assert_has_content()
                try:
                    lexer = get_lexer_by_name(self.arguments[0])
                except ValueError:
                    # no lexer found - use the text one instead of an exception
                    lexer = TextLexer()
                # take an arbitrary option if more than one is given
                formatter = self.options and VARIANTS[self.options.keys()[0]] or DEFAULT
                parsed = highlight(u'\n'.join(self.content), lexer, formatter)
                return [nodes.raw('', parsed, format='html')]

        directives.register_directive('sourcecode', Pygments)
        return publish_parts



# Copyright 2012 posativ <info@posativ.org>. All rights reserved.
# License: BSD Style, 2 clauses. see acrylamid/__init__.py

from os.path import exists

from acrylamid.views import View
from acrylamid.utils import expand, union, joinurl, event
from acrylamid.errors import AcrylamidException

class Entry(View):
    """Creates single full-length entry."""

    def generate(self, request):

        tt = self.env.jinja2.get_template('main.html')

        entrylist = request['entrylist']
        pathes = dict()

        for entry in entrylist:
            if entry.permalink != expand(self.path, entry):
                p = joinurl(self.conf['output_dir'], entry.permalink)
            else:
                p = joinurl(self.conf['output_dir'], expand(self.path, entry))

            if not filter(lambda e: p.endswith(e), ['.xml', '.html']):
                p = joinurl(p, 'index.html')

            if p in pathes:
                raise AcrylamidException("title collision %r in %r" % (entry.permalink,
                                                                       entry.filename))
            pathes[p] = entry

        for path, entry in pathes.iteritems():
            if exists(path) and not entry.has_changed and not tt.has_changed:
                event.skip(path)
                continue

            html = tt.render(env=union(self.env, entrylist=[entry], type='entry'),
                             conf=self.conf, entry=entry)

            yield html, path



# Copyright 2012 posativ <info@posativ.org>. All rights reserved.
# License: BSD Style, 2 clauses. see acrylamid/__init__.py
# -*- encoding: utf-8 -*-

from datetime import datetime
from os.path import exists

from acrylamid.views import View
from acrylamid.utils import joinurl, event, union


class Feed(View):

    def generate(self, request):

        entrylist = filter(lambda e: not e.draft, request['entrylist'])[:self.num_entries]
        tt = self.env.jinja2.get_template('%s.xml' % self.__class__.__name__.lower())

        path = joinurl(self.conf['output_dir'], self.path)
        if not filter(lambda e: path.endswith(e), ['.xml', '.html']):
            path = joinurl(path, 'index.html')


        if exists(path) and not filter(lambda e: e.has_changed, entrylist):
            if not tt.has_changed:
                event.skip(path)
                raise StopIteration

        updated=entrylist[0].date if entrylist else datetime.now()
        html = tt.render(conf=self.conf, env=union(self.env,
                         updated=updated, entrylist=entrylist))

        yield html, path


class Atom(Feed):

    def init(self, num_entries=25):
        self.num_entries = num_entries


class RSS(Feed):

    def init(self, num_entries=25):

        from wsgiref.handlers import format_date_time
        from time import mktime

        self.num_entries = num_entries
        self.env.jinja2.filters['rfc822'] = lambda x: format_date_time(mktime(x.timetuple()))



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#
# Copyright 2012 posativ <info@posativ.org>. All rights reserved.
# License: BSD Style, 2 clauses. see acrylamid/__init__.py

import sys
import os
import glob
import fnmatch
import copy

from acrylamid import log
from acrylamid.errors import AcrylamidException


def get_filters():

    global callbacks
    return callbacks


def index_filters(module, conf, env):
    """Goes through the modules' contents and indexes all classes derieved
    from Filter and have a `match` attribute.

    :param module: the module to index
    :param conf: user config
    :param env: environment
    """

    global callbacks

    cs = [getattr(module, c) for c in dir(module) if not c.startswith('_')]
    for mem in cs:
        if isinstance(mem, meta) and hasattr(mem, 'match'):
            mem.init(conf, env)
            callbacks.append(mem)


def initialize(ext_dir, conf, env, include=[], exclude=[]):
    """Imports and initializes extensions from the directories in the list
    specified by 'ext_dir'.  If no such list exists, the we don't load any
    plugins. 'include' and 'exclude' may contain a list of shell patterns
    used for fnmatch. If empty, this filter is not applied.

    :param ext_dir: list of directories
    :param conf: user config
    :param env: environment
    :param include: a list of filename patterns to include
    :param exclude: a list of filename patterns to exclude
    """

    def get_name(filename):
        """Takes a filename and returns the module name from the filename."""
        return os.path.splitext(os.path.split(filename)[1])[0]

    def get_extension_list(ext_dir, include, exclude):
        """Load all plugins that matches include/exclude pattern in
        given lust of directories.  Arguments as in initializes(**kw)."""

        def pattern(name, incl, excl):
            for i in incl:
                if fnmatch.fnmatch(name, i):
                    return True
            for e in excl:
                if fnmatch.fnmatch(name, e):
                    return False
            if not incl:
                return True
            else:
                return False

        ext_list = []
        for mem in ext_dir:
            files = glob.glob(os.path.join(mem, "*.py"))
            files = [get_name(f) for f in files
                                    if pattern(get_name(f), include, exclude)]
            ext_list += files

        return sorted(ext_list)

    global plugins

    exclude.extend(['mdx_*', 'rstx_*'])
    ext_dir.extend([os.path.dirname(__file__)])

    for mem in ext_dir[:]:
        if os.path.isdir(mem):
            sys.path.insert(0, mem)
        else:
            ext_dir.remove(mem)
            log.error("Filter directory '%s' does not exist. -- skipping" % mem)

    ext_list = get_extension_list(ext_dir, include, exclude)
    for mem in ext_list:
        try:
            _module = __import__(mem)
        except (ImportError, Exception), e:
            log.warn('%r %s: %s', mem, e.__class__.__name__, e)
            continue

        index_filters(_module, conf, env)


class RegexList(list):

    def __contains__(self, other):
        for value in self:
            if isinstance(value, basestring):
                if value == other:
                    return True
            else:
                if value.match(other):
                    return True
        else:
            return False


class meta(type):

    def __init__(cls, name, bases, dct):

        super(meta, cls).__init__(name, bases, dct)
        setattr(cls, 'init', classmethod(dct.get('init', lambda s, x, y: None)))

        if 'match' in dct:
            setattr(cls, 'match', RegexList(dct['match']))


class Filter(object):

    __metaclass__ = meta
    priority = 50.0
    conflicts = []

    def __init__(self, fname, *args):

        self.name = fname
        self.args = args

    def __repr__(self):
        return "<%s %2.f:%s>" % (self.__class__.__name__,
                                 self.priority, self.name)

    def __hash__(self):
        return hash(self.name + repr(self.args))

    def __eq__(self, other):
        return True if hash(other) == hash(self) else False

    def transform(self, text, entry, *args):
        raise NotImplemented


class FilterList(list):
    """a list containing tuples of (filtername, filterfunc, arguments).

    >>> x = FilterList()
    >>> x.append(MyFilter)
    >>> f = x[MyFilter]
    """

    def __contains__(self, y):
        """first checks, wether the item itself is in the list. Next, all Filters
        providing `conflicts` are checked wether y conflicts with a filter.
        Otherwise y is not in FilterList.
        """
        for x in self:
            if x.__class__.__name__ == y.__class__.__name__:
                return True
        for f in y.conflicts:
            for x in self:
                if f in x.match:
                    return True
        return False

    def __getitem__(self, item):

        try:
            f = filter(lambda x: item in x.match, self)[0]
        except IndexError:
            raise ValueError('%s is not in list' % item)

        return f


callbacks = FilterList()



# -*- encoding: utf-8 -*-
#
# Copyright 2012 posativ <info@posativ.org>. All rights reserved.
# License: BSD Style, 2 clauses. see acrylamid/__init__.py

import sys
import os
import time
import locale
import codecs
import tempfile
import shlex

from os.path import join, dirname, getmtime, isfile
from fnmatch import fnmatch
from urlparse import urlsplit
from datetime import datetime
from jinja2 import Environment, FileSystemBytecodeCache

from acrylamid import filters, views, log, utils
from acrylamid.lib.importer import fetch, parse, build
from acrylamid.errors import AcrylamidException
from acrylamid.utils import cache, ExtendedFileSystemLoader, FileEntry, event, escape, \
                            system, filelist

from acrylamid.filters import FilterList


def initialize(conf, env):
    """Initializes Jinja2 environment, prepares locale and configure
    some minor things. Filter and View are inited with conf and env,
    a request dict is returned.
    """
    # set up templating environment
    env['jinja2'] = Environment(loader=ExtendedFileSystemLoader(conf['layout_dir']),
                                bytecode_cache=FileSystemBytecodeCache('.cache/'))
    env['jinja2'].filters.update({'safeslug': utils.safeslug, 'tagify': lambda x: x})

    # initialize the locale, will silently fail if locale is not
    # available and uses system's locale
    try:
        locale.setlocale(locale.LC_ALL, conf.get('lang', ''))
    except (locale.Error, TypeError, KeyError):
        if conf['lang'] in locale.locale_alias:
            locale.setlocale(locale.LC_ALL, locale.locale_alias[conf['lang']])
        else:
            locale.setlocale(locale.LC_ALL, '')
            log.debug("locale set to '%s'", locale.getlocale()[0])
    conf['lang'] = locale.getlocale()[0][:2]

    if 'www_root' not in conf:
        log.warn('no `www_root` specified, using localhost:8000')
        conf['www_root'] = 'http://localhost:8000/'

    env['protocol'], env['netloc'], env['path'], x, y = urlsplit(conf['www_root'])

    # take off the trailing slash for www_root and path
    conf['www_root'] = conf['www_root'].rstrip('/')
    env['path'] = env['path'].rstrip('/')

    # check if encoding is available
    try:
        codecs.lookup(conf['encoding'])
    except LookupError:
        raise AcrylamidException('no such encoding available: %r' % conf['encoding'])

    # XXX implement _optional_ config argments like cache_dir
    # init to conf['cache_dir'] (defaults to '.cache/')
    cache.init()

    # import and initialize plugins
    # XXX put them into defaults
    filters.initialize(conf.get("filters_dir", []), conf, env,
                          exclude=conf.get("filters_ignore", []),
                          include=conf.get("filters_include", []))
    views.initialize(conf.get("views_dir", []), conf, env)
    env['views'] = dict([(v.view, v) for v in views.get_views()])

    return {'conf': conf, 'env': env}


def compile(conf, env, force=False, **options):

    # time measurement
    ctime = time.time()

    # populate env and corrects some conf things
    request = initialize(conf, env)

    if force:
        # acrylamid compile -f
        cache.clear()

    # list of FileEntry-objects reverse sorted by date.
    entrylist = sorted([FileEntry(e, conf) for e in filelist(conf['entries_dir'],
                                                             conf.get('entries_ignore', []))],
                       key=lambda k: k.date, reverse=True)

    # here we store all possible filter configurations
    ns = set()

    # get available filter list, something like with obj.get-function
    # list = [<class head_offset.Headoffset at 0x1014882c0>, <class html.HTML at 0x101488328>,...]
    aflist = filters.get_filters()

    # ... and get all configured views
    _views = views.get_views()

    # filters found in all entries, views and conf.py
    found = sum((x.filters for x in entrylist+_views), []) + request['conf']['filters']

    for val in found:
        # first we for `no` and get the function name and arguments
        f = val[2:] if val.startswith('no') else val
        fname, fargs = f.split('+')[:1][0], f.split('+')[1:]

        try:
            # initialize the filter with its function name and arguments
            fx = aflist[fname](val, *fargs)
            if val.startswith('no'):
                fx.transform = lambda x, y, *z: x
                fx.__hash__ = lambda : 0
        except ValueError:
            try:
                fx = aflist[val.split('+')[:1][0]](val, *fargs)
            except ValueError:
                raise AcrylamidException('no such filter: %s' % val)

        ns.add(fx)

    for entry in entrylist:
        for v in _views:

            # a list that sorts out conflicting and duplicated filters
            flst = FilterList()

            # filters found in this specific entry plus views and conf.py
            found = entry.filters + v.filters + request['conf']['filters']

            for fn in found:
                fx = filter(lambda k: fn == k.name, ns)[0]
                if fx not in flst:
                    flst.append(fx)

            # sort them ascending because we will pop within filters.add
            entry.filters.add(sorted(flst, key=lambda k: (-k.priority, k.name)),
                              context=v.__class__.__name__)

    # lets offer a last break to populate tags or so
    # XXX this API component needs a review
    for v in _views:
        env = v.context(env, {'entrylist': filter(v.condition, entrylist)})

    # now teh real thing!
    for v in _views:

        # XXX the entry should automatically determine its caller (using
        # some sys magic to recursively check wether the calling class is
        # derieved from `View`.)
        for entry in entrylist:
            entry.context = v.__class__.__name__

        request['entrylist'] = filter(v.condition, entrylist)
        tt = time.time()

        for html, path in v.generate(request):
            utils.mkfile(html, path, time.time()-tt, **options)
            tt = time.time()

    log.info('Blog compiled in %.2fs' % (time.time() - ctime))


def autocompile(conf, env, **options):
    """Subcommand: autocompile -- automatically re-compiles when something in
    content-dir has changed and parallel serving files."""

    mtime = -1

    while True:
        ntime = max(getmtime(e) for e in filelist(conf['entries_dir']))
        if mtime != ntime:
            try:
                compile(conf, env, **options)
            except AcrylamidException as e:
                log.fatal(e.message)
                pass
            mtime = ntime
        time.sleep(1)


def clean(conf, everything=False, dryrun=False, **kwargs):
    """Attention: this function may eat your data!  Every create, changed
    or skip event call tracks automatically files. After generation,
    ``acrylamid clean`` will call this function and remove untracked files.

    - with OUTPUT_IGNORE you can specify a list of patterns which are ignored.
    - you can use --dry-run to see what would have been removed
    - by default acrylamid does NOT call this function
    - it removes silently every empty directory

    :param conf: user configuration
    :param every: remove all tracked files, too
    :param dryrun: don't delete, just show what would have been done
    """

    def excluded(path, excl_files):
        """test if path should be ignored"""
        if filter(lambda p: fnmatch(path, p), excl_files):
            return True
        return False

    _tracked_files = utils.get_tracked_files()
    ignored = [join(conf['output_dir'], p) for p in conf['output_ignore']]

    for root, dirs, files in os.walk(conf['output_dir'], topdown=True):
        found = set([join(root, p) for p in files
                     if not excluded(join(root, p), ignored)])
        for i, p in enumerate(found.difference(_tracked_files)):
            if not dryrun:
                os.remove(p)
            event.removed(p)

        if everything:
            for i, p in enumerate(found):
                if not dryrun:
                    os.remove(p)
                event.removed(p)

        # don't visit excluded dirs
        for dir in dirs[:]:
            p = join(root, dir)
            if excluded(p, ignored) or excluded(p+'/', ignored):
                dirs.remove(dir)

    # remove empty directories
    for root, dirs, files in os.walk(conf['output_dir'], topdown=True):
        for p in (join(root, k) for k in dirs):
            try:
                os.rmdir(p)
            except OSError:
                pass


def new(conf, env, title, prompt=True):
    """Subcommand: new -- create a new blog entry the easy way.  Either run
    ``acrylamid new My fresh new Entry`` or interactively via ``acrylamid new``
    and the file will be created using the preferred permalink format."""

    fd, tmp = tempfile.mkstemp(suffix='.txt', dir='.cache/')
    editor = os.getenv('VISUAL') if os.getenv('VISUAL') else os.getenv('EDITOR')

    if not title:
        title = raw_input("Entry's title: ")
    title = escape(title)

    with os.fdopen(fd, 'wb') as f:
        f.write('---\n')
        f.write('title: %s\n' % title)
        f.write('date: %s\n' % datetime.now().strftime(conf['date_format']))
        f.write('---\n\n')

    entry = FileEntry(tmp, conf)
    p = join(conf['entries_dir'], dirname(entry.permalink)[1:])

    try:
        os.makedirs(p.rsplit('/', 1)[0])
    except OSError:
        pass

    filepath = p + '.txt'
    if isfile(filepath):
        raise AcrylamidException('Entry already exists %r' % filepath)
    os.rename(tmp, filepath)
    event.create(filepath)

    if datetime.now().hour == 23 and datetime.now().minute > 45:
        log.info("notice  consider editing entry.date-day after you passed mignight!")

    if not prompt:
        return

    try:
        if editor:
            retcode = os.system(editor + ' ' + filepath)
        elif sys.platform == 'darwin':
            retcode = os.system('open ' + filepath)
        else:
            retcode = os.system('xdg-open ' + filepath)
    except OSError:
        raise AcrylamidException('Could not launch an editor')

    # XXX process detaches... m(
    if retcode < 0:
        raise AcrylamidException('Child was terminated by signal %i' % -retcode)

    if os.stat(filepath)[6] == 0:
        raise AcrylamidException('File is empty!')


def importer(conf, env, url, **options):
    """Subcommand: import -- import entries and settings from an existing RSS/Atom
    feed.  ``acrylamid import http://example.com/feed/`` should do the job.

    If ``pandoc`` or ``html2text`` are available, first pandoc and second html2text
    are used to convert HTML back to Markdown-compatible text.  If you like reST more
    than Markdown, specify ``--format=rst`` and be sure you have either ``pandoc`` or
    ``html2rest`` installed on your system.

    If you don't like any reconversion, simply use ``--format=html``."""

    content = fetch(url, auth=options.get('auth', None))
    defaults, items = parse(content)
    build(conf, env, defaults, items, fmt=options['import_fmt'], keep=options['keep_links'])


def deploy(conf, env, task, *args):
    """Subcommand: deploy -- run the shell command specified in DEPLOYMENT[task] using
    Popen. Use ``%s`` inside your command to let acrylamid substitute ``%s`` with the
    output path, if no ``%s`` is set, the path is appended  as first argument. Every
    argument after ``acrylamid deploy task ARG1 ARG2``."""

    cmd = shlex.split(conf.get('deployment', {}).get(task, None))
    if not cmd:
        raise AcrylamidException('no tasks named %r in conf.py' % task)

    try:
        cmd[cmd.index('%s')] = conf['output_dir']
    except ValueError:
        cmd.append(conf['output_dir'])

    # apply ARG1 ARG2 ... and -v --long-args to the command, e.g.:
    # $> acrylamid deploy task arg1 -- -b --foo
    cmd.extend(args)

    try:
        result = system(cmd)
        print '\n'.join('    '+line for line in result.strip().split('\n'))
    except OSError as e:
        raise AcrylamidException(e.message)


__all__ = ["compile", "autocompile", "new"]



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#
# Copyright 2011 posativ <info@posativ.org>. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
# The views and conclusions contained in the software and documentation are
# those of the authors and should not be interpreted as representing official
# policies, either expressed or implied, of posativ <info@posativ.org>.

__version__ = "0.3.0"
__author__ = 'posativ <info@posativ.org>'
__url__ = 'https://github.com/posativ/acrylamid/'

import sys
reload(sys)
sys.setdefaultencoding('utf-8')

import os
import time
import traceback
import signal

from optparse import OptionParser, make_option, SUPPRESS_HELP
from textwrap import fill
from acrylamid import defaults, log, commands
from acrylamid.errors import AcrylamidException

signal.signal(signal.SIGINT, signal.SIG_DFL)
sys.path.insert(0, os.path.dirname(__package__))


class Environment(dict):

    def __getattribute__(self, attr):
        try:
            return getattr(super(Environment, self), attr)
        except AttributeError:
            try:
                return self[attr]
            except KeyError:
                raise AttributeError

    def __setattribute__(self, attr, value):
        self[attr] = value


class Acryl:
    """Main class for acrylamid functionality.  It handles initialization,
    defines default behavior, and also pushes the request through all
    steps until the output is rendered and we're complete."""

    def __init__(self, conf=None):
        """Sets configuration and environment and creates the Request
        object"""

        global sys

        usage = "usage: %prog <subcommand> [options] [args]"
        epilog = None

        options, constants = [], [
            make_option("-h", "--help", action="store_true", help=SUPPRESS_HELP),
            make_option("-v", "--verbose", action="store_const", dest="verbosity",
                        help=SUPPRESS_HELP, const=log.SKIP),
            make_option("-q", "--quit", action="store_const", dest="verbosity",
                        help=SUPPRESS_HELP, const=log.WARN, default=log.INFO),
            make_option("-C", "--no-color", action="store_false", dest="colors",
                        default=True, help=SUPPRESS_HELP),
            make_option("--version", action="store_true", dest="version",
                            help=SUPPRESS_HELP, default=False)
            ]

        # reorder `prog help foo` and `prog --help foo`
        if len(sys.argv) > 2 and sys.argv[1] in ('--help', 'help'):
            sys.argv[1], sys.argv[2] = sys.argv[2], '--help'

        if len(sys.argv) <= 1 or sys.argv[1] in ('-h', '--help', 'help'):
            options, constants = [], [
                make_option("-q", "--quiet", action="store_const", dest="verbosity",
                        help="less verbose", const=log.WARN, default=log.INFO),
                make_option("-v", "--verbose", action="store_const", dest="verbosity",
                            help="more verbose", const=log.SKIP),
                make_option("-C", "--no-color", action="store_false", dest="colors",
                            default=True, help="disable color"),
                make_option("-h", "--help", action="store_true",
                            help="show this help message and exit"),
                make_option("--version", action="store_true", dest="version",
                            help="print version details", default=False)
            ]
            epilog = ("Commands:\n"
                      "  init           initializes base structure in DIR\n"
                      "  create  (new)  creates a new entry\n"
                      "  compile (co)   compile blog\n"
                      "  view           fire up built-in webserver\n"
                      "  autocompile    automatic compilation and serving (short aco)\n"
                      "  clean   (rm)   remove abandoned files\n"
                      "  import         import content from URL\n"
                      "  deploy         run a given TASK\n"
                      "\nAll subcommands except `init` require a conf.py file.\n")

        # --- init params --- #
        elif sys.argv[1] in ('init', 'initialize'):
            usage = "%prog " + sys.argv[1] + " [DEST|FILE] [-p]"
            options = [
                make_option("-f", "--force", action="store_true", dest="force",
                        help="don't ask, just overwrite", default=False),
                make_option("--xhtml", action="store_const", dest="theme", const="xhtml",
                            help="use XHTML theme", default="html5"),
                make_option("--html5", action="store_const", dest="theme", const="html5",
                            help="use HTML5 theme")
                ]
        # --- init params --- #
        elif sys.argv[1] in ('new',):
            usage = "%prog " + sys.argv[1] + " [args]"
            epilog = ("Takes all leading [args] as title or prompt if none given. creates "
                      "a new entry based on your PERMALINK_FORMAT and opens it with your "
                      "favourite EDITOR.")
            epilog = fill(epilog)+'\n'
        # --- gen params --- #
        elif sys.argv[1] in ('compile', 'co', 'generate', 'gen'):
            usage = "%prog " + sys.argv[1] + " [-fn]"
            options = [
                make_option("-f", "--force", action="store_true", dest="force",
                            help="clear cache before compilation", default=False),
                make_option("-n", "--dry-run", dest="dryrun", action='store_true',
                            default=False, help="show what would have been compiled"),
                make_option("-i", "--ignore", dest="ignore", action="store_true",
                            default=False, help="ignore critical errors")
            ]
        # --- webserver params --- #
        elif sys.argv[1] in ('view', 'serve', 'srv'):
            usage = "%prog " + sys.argv[1] + " [-p]"
            options = [
                make_option("-p", "--port", dest="port", type=int, default=8000,
                            help="webserver port"),
            ]
        # --- autocompile params --- #
        elif sys.argv[1] in ('autocompile', 'aco'):
            usage = "%prog " + sys.argv[1] + " [-pf]"
            options = [
                make_option("-f", "--force", action="store_true", dest="force",
                            help="clear cache before compilation", default=False),
                make_option("-i", "--ignore", dest="ignore", action="store_true",
                            default=False, help="ignore critical errors"),
                make_option("-p", "--port", dest="port", type=int, default=8000,
                            help="webserver port"),
            ]
        # --- clean params --- #
        elif sys.argv[1] in ('clean', 'rm'):
            usage = "%prog " + sys.argv[1] + " [-fn]"
            options = [
                make_option("-f", "--force", action="store_true", dest="force",
                            help="remove all files generated by Acrylamid", default=False),
                make_option("-n", "--dry-run", dest="dryrun", action='store_true',
                            default=False, help="show what would have been deleted"),
            ]
        # --- import params --- #
        elif sys.argv[1] in ('import', ):
            usage = "%prog " + sys.argv[1] + " [-mk]"
            options = [
                make_option("-m", dest="import_fmt", default="Markdown",
                            help="reconvert HTML to MARKUP"),
                make_option("-k", "--keep-links", dest="keep_links", action="store_true",
                            help="keep permanent links"),
            ]

        class AcrylParser(OptionParser):
            # -- http://stackoverflow.com/q/1857346
            def format_epilog(self, formatter):
                if epilog is None:
                    return ''
                return '\n' + self.epilog

        parser = AcrylParser(option_list=options+constants, usage=usage,
                              add_help_option=False, epilog=epilog)
        (options, args) = parser.parse_args()

        if len(sys.argv) <= 1 or sys.argv[1] == 'help' or options.help:
            parser.print_help()
            sys.exit(0)

        # initialize colored logger
        log.init('acrylamid', level=options.verbosity, colors=options.colors)

        env = Environment({'version': __version__, 'author': __author__, 'url': __url__})
        env['options'] = options

        if options.version:
            print 'acrylamid ' + env.version
            sys.exit(0)

        # -- init -- #
        # TODO: acrylamid init --layout_dir=somedir to overwrite defaults

        if 'init' in args:
            if len(args) == 2:
                defaults.init(args[1], options.theme, options.force)
            else:
                defaults.init('.', options.theme, options.force)
            sys.exit(0)

        # -- teh real thing -- #
        conf = defaults.conf

        try:
            ns = {}
            execfile('conf.py', ns)
            conf.update(dict([(k.lower(), ns[k]) for k in ns if k.upper() == k]))
        except OSError:
            log.critical('no config file found: %s. Try "acrylamid init".', options.conf)
            sys.exit(1)
        except Exception, e:
            log.critical("%s in `conf.py`" % e.__class__.__name__)
            traceback.print_exc(file=sys.stdout)
            sys.exit(1)

        conf['output_dir'] = conf.get('output_dir', 'output/')
        conf['entries_dir'] = conf.get('entries_dir', 'content/')
        conf['layout_dir'] = conf.get('layout_dir', 'layouts/')

        assert defaults.check_conf(conf)
        conf.update(dict((k, v) for k, v in options.__dict__.iteritems() if v != None))

        # -- run -- #

        if args[0] in ('gen', 'generate', 'co', 'compile'):
            log.setLevel(options.verbosity-5)
            try:
                commands.compile(conf, env, **options.__dict__)
            except AcrylamidException as e:
                log.fatal(e.message)
                sys.exit(1)

        elif args[0] in ('new', 'create'):
            try:
                commands.new(conf, env, title=' '.join(args[1:]), prompt=log.level()<log.WARN)
            except AcrylamidException as e:
                log.fatal(e.message)
                sys.exit(1)

        elif args[0] in ('clean', 'rm'):
            try:
                log.setLevel(options.verbosity+5)
                options.ignore = True  # we don't bother the user here
                commands.compile(conf, env, dryrun=True, force=False)
                log.setLevel(options.verbosity)
                commands.clean(conf, everything=options.force, **options.__dict__)
            except AcrylamidException as e:
                log.fatal(e.message)
                sys.exit(1)

        elif args[0] in ('srv', 'serve', 'view'):
            from acrylamid.lib.httpd import Webserver
            ws = Webserver(options.port, conf['output_dir']); ws.start()
            log.info(' * Running on http://127.0.0.1:%i/' % options.port)

            try:
                while True:
                    time.sleep(1)
            except (SystemExit, KeyboardInterrupt, Exception) as e:
                ws.kill_received = True
                sys.exit(0)

        elif args[0] in ('aco', 'autocompile'):
            from acrylamid.lib.httpd import Webserver
            # XXX compile on request _or_ use inotify/fsevent
            ws = Webserver(options.port, conf['output_dir']); ws.start()
            log.info(' * Running on http://127.0.0.1:%i/' % options.port)

            try:
                commands.autocompile(conf, env, **options.__dict__)
            except (SystemExit, KeyboardInterrupt, Exception) as e:
                ws.kill_received = True
                log.error(e.message)
                traceback.print_exc(file=sys.stdout)
                sys.exit(0)

        elif args[0] in ('import', ):
            if len(args) <= 1:
                log.fatal('import requires a valid URL')
                sys.exit(1)
            try:
                commands.importer(conf, env, args[1], **options.__dict__)
            except AcrylamidException as e:
                log.critical(e.message)
                sys.exit(1)
        elif args[0] in ('deploy', 'dp', 'task'):
            if len(args) <= 1:
                log.fatal('deploy requires a build TASK')
                sys.exit(1)
            try:
                commands.deploy(conf, env, args[1], *args[2:])
            except AcrylamidException as e:
                log.critical(e.message)
                sys.exit(1)
        else:
            log.critical('No such command!')
            sys.exit(2)



# License: BSD Style, 2 clauses. see acrylamid/__init__.py
# -*- encoding: utf-8 -*-

from acrylamid.filters import Filter
from acrylamid.filters import log
from acrylamid.utils import cached_property

import re
import os
import codecs
from HTMLParser import HTMLParser, HTMLParseError
from cgi import escape
from os.path import join, dirname, basename


class HyphenPatternNotFound(Exception):
    pass


class Hyphenator:
    """ Hyphenation, using Frank Liang's algorithm.

    This module provides a single function to hyphenate words.  hyphenate_word takes
    a string (the word), and returns a list of parts that can be separated by hyphens.

    >>> hyphenate_word("hyphenation")
    ['hy', 'phen', 'ation']
    >>> hyphenate_word("supercalifragilisticexpialidocious")
    ['su', 'per', 'cal', 'ifrag', 'ilis', 'tic', 'ex', 'pi', 'ali', 'do', 'cious']
    >>> hyphenate_word("project")
    ['project']

    Ned Batchelder, July 2007.
    This Python code is in the public domain."""

    __version__ = '1.0.20070709'

    def __init__(self, chars, patterns, exceptions=''):
        self.chars = unicode('[.' + chars + ']')
        self.tree = {}
        for pattern in patterns.split():
            self._insert_pattern(pattern)

        self.exceptions = {}
        for ex in exceptions.split():
            # Convert the hyphenated pattern into a point array for use later.
            self.exceptions[ex.replace('-', '')] = [0] + [int(h == '-') for h in re.split(r"[a-z]", ex)]

    def _insert_pattern(self, pattern):
        # Convert the a pattern like 'a1bc3d4' into a string of chars 'abcd'
        # and a list of points [ 1, 0, 3, 4 ].
        chars = re.sub('[0-9]', '', pattern)
        points = [int(d or 0) for d in re.split(self.chars, pattern)]

        # Insert the pattern into the tree.  Each character finds a dict
        # another level down in the tree, and leaf nodes have the list of
        # points.
        t = self.tree
        for c in chars:
            if c not in t:
                t[c] = {}
            t = t[c]
        t[None] = points

    def hyphenate_word(self, word):
        """ Given a word, returns a list of pieces, broken at the possible
            hyphenation points.
        """
        # Short words aren't hyphenated.
        if len(word) <= 4:
            return [word]
        # If the word is an exception, get the stored points.
        if word.lower() in self.exceptions:
            points = self.exceptions[word.lower()]
        else:
            work = '.' + word.lower() + '.'
            points = [0] * (len(work) + 1)
            for i in range(len(work)):
                t = self.tree
                for c in work[i:]:
                    if c in t:
                        t = t[c]
                        if None in t:
                            p = t[None]
                            for j in range(len(p)):
                                points[i + j] = max(points[i + j], p[j])
                    else:
                        break
            # No hyphens in the first two chars or the last two.
            points[1] = points[2] = points[-2] = points[-3] = 0

        # Examine the points to build the pieces list.
        pieces = ['']
        for c, p in zip(word, points[2:]):
            pieces[-1] += c
            if p % 2:
                pieces.append('')
        return pieces


class Separator(HTMLParser):
    """helper class to apply Hyphenator to each word except in pre, code,
    math and em tags."""

    def __init__(self, html, hyphenationfunc, length=10):
        HTMLParser.__init__(self)
        self.hyphenate = hyphenationfunc
        self.length = length
        self.result = []
        self.stack = []

        self.feed(html)

    def handle_starttag(self, tag, attrs):
        """Apply and stack each read tag until we reach maxword."""

        def tagify(tag, attrs):
            """convert parsed tag back into a html tag"""
            if attrs:
                return '<%s %s>' % (tag, ' '.join(['%s="%s"' % (k, escape(v))
                                        for k, v in attrs]))
            else:
                return '<%s>' % tag

        self.stack.append(tag)
        self.result.append(tagify(tag, attrs))

    def handle_data(self, data):
        """Hyphenate words longer than 10 characters."""

        if filter(lambda i: i in self.stack, ['pre', 'code', 'math', 'em']):
            pass
        else:
            split = [word for word in re.split(r"[.:,\s!?+=\(\)/-]+", data)
                     if len(word) > self.length]
            for word in split:
                hyphenated = '&shy;'.join(self.hyphenate(word))
                data = data.replace(word, hyphenated)

        self.result.append(data)

    def handle_endtag(self, tag):
        """Until we reach not the maxwords limit, we can safely pop every ending tag,
           added by handle_starttag. Afterwards, we apply missing endings tags if missing."""
        try:
            self.stack.pop()
        except IndexError:
            pass
        self.result.append('</%s>' % tag)

    def handle_entityref(self, name):
        self.result.append('&' + name + ';')

    def handle_charref(self, char):
        self.result.append('&#' + char + ';')


def build(lang):
    """build the Hyphenator from given language.  If you want add more, see
    http://tug.org/svn/texhyphen/trunk/hyph-utf8/tex/generic/hyph-utf8/patterns/txt/ ."""

    def gethyph(lang, directory='hyph/', prefix='hyph-'):

        for la in [prefix + lang, prefix + lang[:2]]:
            for p in os.listdir(directory):
                f = os.path.basename(p)
                if f.startswith(la):
                    return join(directory, p)
        else:
            raise HyphenPatternNotFound("no hyph-definition found for '%s'" % lang)

    dir = os.path.join(dirname(__file__), 'hyph/')
    fpath = gethyph(lang, dir).rsplit('.', 2)[0]
    try:
        with codecs.open(fpath + '.chr.txt', encoding='utf-8') as f:
            chars = ''.join([line[0] for line in f.readlines()])
        with codecs.open(fpath + '.pat.txt', encoding='utf-8') as f:
            patterns = f.read()
    except IOError:
        raise HyphenPatternNotFound('hyph/%s.chr.txt or hyph/%s.pat.txt missing' % (lang, lang))

    hyphenator = Hyphenator(chars, patterns, exceptions='')
    del patterns
    del chars
    log.debug("built Hyphenator from <%s>" % basename(fpath))
    return hyphenator.hyphenate_word


class Hyphenate(Filter):

    match = [re.compile('^(H|h)yph')]

    @cached_property
    def default(self):
        try:
            # build default hyphenate_word using conf's lang (if available)
            return build(self.conf['lang'].replace('_', '-'))
        except HyphenPatternNotFound as e:
            log.warn(e.message)
            return lambda x: [x]

    def init(self, conf, env):
        self.conf = conf

    def transform(self, content, entry, *args):
        if entry.lang != self.conf['lang']:
            try:
                hyphenate_word = build(entry.lang.replace('_', '-'))
            except HyphenPatternNotFound as e:
                log.once(warn=e.message)
                hyphenate_word = lambda x: [x]
        else:
            hyphenate_word = self.default

        try:
            length = int(args[0])
        except (ValueError, IndexError) as e:
            if e.__class__.__name__ == 'ValueError':
                log.warn('Hyphenate: invalid length argument %r', args[0])
            length = 10

        try:
            return ''.join(Separator(content, hyphenate_word, length=length).result)
        except HTMLParseError as e:
            log.warn('%s: %s in %s' % (e.__class__.__name__, e.msg, entry.filename))
            return content
